/*
 * Copyright (c) 2022 Libre Solar Technologies GmbH
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/uart.h>

#include <string.h>
#include <time.h>
#include <zephyr/sys/timeutil.h>

//#include <SEGGER_RTT.h>

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(gps);
//LOG_MODULE_REGISTER(gps, LOG_LEVEL_DBG);

#include "lwgps.h"
#include "gps.h"
#include "calendar.h"

#define MY_STACK_SIZE 2048
#define MY_PRIORITY 5

K_FIFO_DEFINE(fifo_gps_sat_data);
K_FIFO_DEFINE(fifo_gps_data);
K_SEM_DEFINE(gps_data_tosend, 0, 1);
K_SEM_DEFINE(gps_init_ok, 0, 1);

/*#define RTT_BUFFER_DOWN_SIZE 512
#define RTT_BUFFER_UP_SIZE 512

struct rtt_config_t {
	void *up_buffer;
	size_t up_size;
	void *down_buffer;
	size_t down_size;
	uint8_t channel;
};*/



/* change this to any other UART peripheral if desired */
#define UART_DEVICE_NODE DT_NODELABEL(uart0)

#define MSG_SIZE 128

/* queue to store up to 10 messages (aligned to 4-byte boundary) */
K_MSGQ_DEFINE(uart_msgq, MSG_SIZE, 10, 4);

static K_SEM_DEFINE(segger_init_ok, 0, 1);

static const struct device *const uart_dev = DEVICE_DT_GET(UART_DEVICE_NODE);

static char nmea_sentence[MSG_SIZE];
static int nmea_position = 0;
static bool gps_state = false; 

const unsigned char ubxNMEA[] = { 0x06, 0x17, 0x14, 0x00, 0x00, 0x41, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
const unsigned char ubxGNSS[] = { 0x06, 0x3E, 0x3C, 0x00, 0x00, 0x20, 0x20, 0x07, 0x00, 0x08, 0x10, 0x00, 0x01, 0x00, 0x01, 0x01, 0x01, 0x01, 0x03, 0x00, 0x01, 0x00, 0x01, 0x01, 0x02, 0x04, 0x08, 0x00, 0x01, 0x00, 0x01, 0x01, 0x03, 0x08, 0x10, 0x00, 0x00, 0x00, 0x01, 0x01, 0x04, 0x00, 0x08, 0x00, 0x00, 0x00, 0x01, 0x03, 0x05, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x05, 0x06, 0x08, 0x0E, 0x00, 0x01, 0x00, 0x01, 0x01 };
//const unsigned char ubxPoweroff[] = { 0x02, 0x41, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00 };
const unsigned char ubxEnablePulse[] = { 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x32, 0x00, 0x00, 0x00, 0x40, 0x42, 0x0f, 0x00, 0x40, 0x42, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa0, 0x86, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x77, 0x00, 0x00, 0x00};
const unsigned char ubxDisablePulse[] = { 0x06, 0x31, 0x20, 0x00, 0x00, 0x01, 0x00, 0x00, 0x32, 0x00, 0x00, 0x00, 0x40, 0x42, 0x0f, 0x00, 0x40, 0x42, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x77, 0x00, 0x00, 0x00};
const unsigned char GNSSStart[] = { 0x06,0x04,0x04,0x00,0x00,0x00,0x09,0x00 };
const unsigned char GNSSStop[] = { 0x06,0x04,0x04,0x00,0x00,0x00,0x08,0x00 };
const unsigned char enablePUBX0[] = { 0x06, 0x01, 0x08, 0x00, 0xf1, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00 };
const unsigned char disablePUBX0[] = { 0x06, 0x01, 0x08, 0x00, 0xf1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
const unsigned char enablePUBX4[] = { 0x06, 0x01, 0x08, 0x00, 0xf1, 0x04, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00 };
const unsigned char disablePUBX4[] = { 0x06, 0x01, 0x08, 0x00, 0xf1, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
const unsigned char rate2Hz2Mes[] = { 0x06, 0x08, 0x06, 0x00, 0xf4, 0x02, 0x02, 0x00, 0x00, 0x00 };
const unsigned char rate1Hz1Mes[] = { 0x06, 0x08, 0x06, 0x00, 0xe8, 0x03, 0x01, 0x00, 0x00, 0x00 };
const unsigned char rate1Hz5Mes[] = { 0x06, 0x08, 0x06, 0x00, 0xe8, 0x03, 0x05, 0x00, 0x00, 0x00 };

const char disableRMC[] = "PUBX,40,RMC,0,0,0,0,0,0";
const char disableGLL[] = "PUBX,40,GLL,0,0,0,0,0,0";
const char disableGSV[] = "PUBX,40,GSV,0,0,0,0,0,0";
const char disableGSA[] = "PUBX,40,GSA,0,0,0,0,0,0";
const char disableGGA[] = "PUBX,40,GGA,0,0,0,0,0,0";
const char disableVTG[] = "PUBX,40,VTG,0,0,0,0,0,0";
const char disableZDA[] = "PUBX,40,ZDA,0,0,0,0,0,0";
const char disableGNS[] = "PUBX,40,GNS,0,0,0,0,0,0";
//const char disableGBS[] PROGMEM = "PUBX,40,GBS,0,0,0,0,0,0";
const char disableGST[] = "PUBX,40,GST,0,0,0,0,0,0";

const char enableRMC[] = "PUBX,40,RMC,0,1,0,0,0,0";
const char enableGLL[] = "PUBX,40,GLL,0,1,0,0,0,0";
const char enableGSV[] = "PUBX,40,GSV,0,1,0,0,0,0";
const char enableGSA[] = "PUBX,40,GSA,0,1,0,0,0,0";
const char enableGGA[] = "PUBX,40,GGA,0,1,0,0,0,0";
const char enableVTG[] = "PUBX,40,VTG,0,1,0,0,0,0";
const char enableZDA[] = "PUBX,40,ZDA,0,1,0,0,0,0";
const char enableGNS[] = "PUBX,40,GNS,0,1,0,0,0,0";
//const char enableGBS[] PROGMEM = "PUBX,40,GBS,0,1,0,0,0,0";
const char enableGST[] = "PUBX,40,GST,0,1,0,0,0,0\r\n";

/* void print_segger(char *buf)
{
	int msg_len = strlen(buf);

	for (int i = 0; i < msg_len; i++) {
		//uart_poll_out(uart_dev, buf[i]);
		SEGGER_RTT_Write(1,&buf[i],1);
	}
}

void print_segger_byte(char buf)
{
    SEGGER_RTT_Write(1,&buf,1);
}*/

void serial_cb(const struct device *dev, void *user_data)
{
	uint8_t c;

	if (!uart_irq_update(uart_dev)) {
		return;
	}

	if (!uart_irq_rx_ready(uart_dev)) {
		return;
	}

	/* read until FIFO empty */
	while (uart_fifo_read(uart_dev, &c, 1) == 1) {
		if(c == '$')
		{
			nmea_position = 0;
			nmea_sentence[nmea_position] = c;
			nmea_position++;
		} else if(nmea_position > 0)
		{
			//if((nmea_position == 1 && c != 'G') || (nmea_position == 2 && (c != 'L' && c != 'N' && c != 'P')) || (nmea_position == 6 && c!= ','))
			//{
			//	nmea_position = 0;
			//}
			if(nmea_position < (sizeof(nmea_sentence) -1))
			{
				nmea_sentence[nmea_position] = c;
				nmea_position++;
				if(c == '\n')
				{
					nmea_sentence[nmea_position] = '\0';
					k_msgq_put(&uart_msgq, &nmea_sentence, K_NO_WAIT);
					nmea_position = 0;
				}
			} else {
				printk("overflow\n");
				nmea_position = 0;
			}

		}
		//print_segger_byte(c);
	}
}

static char formatHex( uint8_t val )
{
  val &= 0x0F;
  return (val >= 10) ? ((val - 10) + 'A') : (val + '0');
}

void print_uart_byte(char buf)
{
		uart_poll_out(uart_dev, buf);
}
void print_uart(char *buf)
{
	int msg_len = strlen(buf);

	for (int i = 0; i < msg_len; i++) {
		uart_poll_out(uart_dev, buf[i]);
	}
}

void print_uart_const(const char *buf)
{
	int msg_len = strlen(buf);

	print_uart_byte('$');
	uint8_t crc = 0;
	for (int i = 0; i < msg_len; i++) {
		print_uart_byte(buf[i]);
		crc ^= buf[i];
  }
	print_uart_byte('*');
	print_uart_byte(formatHex(crc>>4));
	print_uart_byte(formatHex(crc));
	print_uart_byte('\r');
	print_uart_byte('\n');
}

void sendUBX( const unsigned char *progmemBytes, size_t len)
{
  /*if(wakeup == true)
  {
    gpsPort.write( 0xFF );
    delay( 500 );
    
  }*/
  print_uart_byte( 0xB5 ); // SYNC1
  print_uart_byte( 0x62 ); // SYNC2
  uint8_t a = 0, b = 0;
  while (len-- > 0) {
    uint8_t c = *(progmemBytes++);
    a += c;
    b += a;
    print_uart_byte( c );
  }

  print_uart_byte( a ); // CHECKSUM A
  print_uart_byte( b ); // CHECKSUM B
} // sendUBX


void init_gps()
{
  sendUBX(GNSSStart, sizeof(GNSSStart));
  k_msleep(100);
  sendUBX(ubxNMEA, sizeof(ubxNMEA));
  k_msleep(100);
  sendUBX(ubxGNSS, sizeof(ubxGNSS));
  k_msleep(100);
  sendUBX(rate1Hz5Mes, sizeof(rate1Hz5Mes));
  k_msleep(100);
  sendUBX(ubxDisablePulse, sizeof(ubxDisablePulse));
  k_msleep(100);

  print_uart_const(disableRMC);
  k_msleep(100);
  print_uart_const(disableGSV);
  k_msleep(100);
  print_uart_const(disableVTG);
  k_msleep(100);
  print_uart_const(disableGLL);
  k_msleep(100);
  print_uart_const(enableGGA);
  k_msleep(100);
  print_uart_const(enableGSA);
  k_msleep(100);
  sendUBX(enablePUBX0, sizeof(enablePUBX0));
  k_msleep(100);
  sendUBX(enablePUBX4, sizeof(enablePUBX4));
}

/*void segger_read_ep(void *, void *, void *)
{

        k_sem_take(&segger_init_ok, K_FOREVER);

        while (1) {
            char tx_buf;
		unsigned numbytes = 0;
		do {
                numbytes = SEGGER_RTT_Read(1, &tx_buf, 1);
                if(numbytes > 0)
                {
                //print_uart("Echo: ");
                //for(int i = numbytes; i < MSG_SIZE; i++)
                //{
                //    tx_buf[i] = '\0';
                //}
                printk("Segger received %d bytes: %d\n", numbytes, tx_buf);
                print_uart_byte(tx_buf);
                //print_uart("\r\n");
		}
		} while(numbytes > 0);
		k_msleep(100);
		k_yield();
                }
}

K_THREAD_DEFINE(segger_read_tid, MY_STACK_SIZE,
                segger_read_ep, NULL, NULL, NULL,
                MY_PRIORITY, 0, 0);
*/

void stop_gps() {
  sendUBX(GNSSStop, sizeof(GNSSStop));
}

void gpsPower(bool state) {
  gps_state = state;
  if(state == true) {
    LOG_WRN("Start GPS");
    init_gps();
  } else {
    LOG_WRN("Stop GPS");
    stop_gps();
  }
}

void gpsBoost(bool state) {
  if(state) {
    sendUBX(rate1Hz1Mes, sizeof(rate1Hz1Mes));
    k_msleep(100);
  } else {
    sendUBX(rate1Hz5Mes, sizeof(rate1Hz5Mes));
    k_msleep(100);
  }
}

void gps_ep(void *, void *, void *)
{
  char tx_buf[MSG_SIZE];

  lwgps_t hgps;

  /*	struct rtt_config_t rtt_config;
  rtt_config.up_size = RTT_BUFFER_UP_SIZE;
  rtt_config.down_size = RTT_BUFFER_DOWN_SIZE;
  rtt_config.up_buffer = k_malloc(RTT_BUFFER_UP_SIZE);
  rtt_config.down_buffer = k_malloc(RTT_BUFFER_DOWN_SIZE);
  rtt_config.channel = 1;

  SEGGER_RTT_ConfigUpBuffer(rtt_config.channel, "RTT GPS", rtt_config.up_buffer, rtt_config.up_size, SEGGER_RTT_MODE_NO_BLOCK_SKIP);
  SEGGER_RTT_ConfigDownBuffer(rtt_config.channel, "RTT GPS", rtt_config.down_buffer, rtt_config.down_size, SEGGER_RTT_MODE_NO_BLOCK_SKIP);

        k_sem_give(&segger_init_ok);*/

  if (!device_is_ready(uart_dev)) {
    printk("UART device not found!");
    return;
  }

  gpsPower(true);
  k_msleep(250);
  gpsPower(false);

  lwgps_init(&hgps);

  //k_fifo_init(&fifo_gps_sat_data);
  //k_sem_init(&gps_data_tosend, 0, 2);

  /* configure interrupt and callback to receive data */
  int ret = uart_irq_callback_user_data_set(uart_dev, serial_cb, NULL);

  if (ret < 0) {
    if (ret == -ENOTSUP) {
      printk("Interrupt-driven UART API support not enabled\n");
    } else if (ret == -ENOSYS) {
      printk("UART device does not support interrupt-driven API\n");
    } else {
      printk("Error setting UART callback: %d\n", ret);
    }
    return;
  }
  uart_irq_rx_enable(uart_dev);
  LOG_INF("Hello GPS");

  k_sem_give(&gps_init_ok);

  uint8_t gnss_count = 0;
  uint32_t altitude = 0;
  uint8_t gps = 0;
  uint8_t glonass = 0;
  uint8_t galileo = 0;
  /* indefinitely wait for input from the user */
  while (k_msgq_get(&uart_msgq, &tx_buf, K_FOREVER) == 0) {
    //print_uart("Echo: ");
    //char talkerid[3];
    //minmea_talker_id(&talkerid, tx_buf);
    //LOG_INF("NMEA msg");
    if(gps_state == true)
    {
      LOG_DBG("NMEA Sentence: %s", tx_buf);
      lwgps_process(&hgps, tx_buf, strlen(tx_buf));
      //LOG_INF("NMEA Sentence %s", hgps.p.term_str);
      if(hgps.p.stat == STAT_GSA) { 

        uint8_t sat_count = 0;
        for(int i = 0; i < sizeof(hgps.satellites_ids); i++)
        {
          if(hgps.satellites_ids[i] != 0) { sat_count++; }
        }
        LOG_INF("systemId: %d %d", hgps.systemId, sat_count);
        if(hgps.systemId == 1) {
          gps = sat_count;
        } else if(hgps.systemId == 2) {
          glonass = sat_count;
        } else if(hgps.systemId == 3) {
          galileo = sat_count;
        }
        gnss_count++;
        if(gnss_count >= 3 && (gps != 0 || glonass != 0 || galileo != 0)) {
          struct gps_sat_data_t *sat_data = k_malloc(sizeof(*sat_data));
          sat_data->gps = gps;
          sat_data->glonass = glonass;
          sat_data->galileo = galileo;
          //LOG_INF("Send Sat Data");
          k_fifo_put(&fifo_gps_sat_data, sat_data);
          k_sem_give(&gps_data_tosend);
          gnss_count = 0;
        }
      } else if(hgps.p.stat == STAT_GGA) { 
        altitude = (int32_t) (hgps.altitude*100);
      } else if(hgps.p.stat == STAT_UBX_POSITION) { 
        LOG_INF("Position %d %d", hgps.latitude, hgps.longitude);
        LOG_INF("Altitude: %d", altitude);
        LOG_INF("DOP %d %d", hgps.dop_h, hgps.dop_v);
        LOG_INF("Accuracy  H:%d V:%d", hgps.acc_h, hgps.acc_v);
        LOG_INF("Speed:%d Course:%d", hgps.speed, hgps.course);
        if(hgps.dop_h != 999 && hgps.dop_v != 999) {
          struct gps_data_t *data =  k_malloc(sizeof(*data));
          data->latitude = hgps.latitude;
          data->longitude = hgps.longitude;
          data->altitude = altitude;
          data->dop_h = hgps.dop_h;
          data->dop_v = hgps.dop_v;
          data->acc_h = hgps.acc_h;
          data->acc_v = hgps.acc_v;
          data->course = hgps.course;
          data->speed = hgps.speed;
          k_fifo_put(&fifo_gps_data, data);
          k_sem_give(&gps_data_tosend);
        }
      } else if(hgps.p.stat == STAT_UBX_TIME) { 
        if(hgps.year >= 24)
        {
          struct tm current_tm;
          current_tm.tm_hour = hgps.hours;
          current_tm.tm_min = hgps.minutes;
          current_tm.tm_sec = hgps.seconds;
          current_tm.tm_mday= hgps.date;
          current_tm.tm_mon = hgps.month-1;
          current_tm.tm_year = hgps.year + 2000 - 1900;
          setTs((uint32_t) timeutil_timegm64(&current_tm));
          sendUBX(disablePUBX4, sizeof(disablePUBX4));
        }
      }
    } else if(strcmp(tx_buf, "$GNTXT,01,01,02,Stopping GNSS*52\r\n") != 0) {
      gpsPower(false);
    }
  }
}
K_THREAD_DEFINE(gps_tid, MY_STACK_SIZE,
                gps_ep, NULL, NULL, NULL,
                MY_PRIORITY, 0, 0);
