/*
 * Copyright (c) 2022 Libre Solar Technologies GmbH
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>

#include <zephyr/sys/byteorder.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/services/bas.h>
#include <math.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/drivers/gpio.h>

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(main);

#include <string.h>

#include "gps.h"
#include "bme280.h"
#include "ruuvi.h"
#include "battery.h"

#define STACKSIZE 2048
#define PRIORITY 7

#define LED0_NODE DT_ALIAS(led0)
static const struct gpio_dt_spec led0 = GPIO_DT_SPEC_GET(LED0_NODE, gpios);
#define LED1_NODE DT_ALIAS(led1)
static const struct gpio_dt_spec led1 = GPIO_DT_SPEC_GET(LED1_NODE, gpios);
#define LED2_NODE DT_ALIAS(led2)
static const struct gpio_dt_spec led2 = GPIO_DT_SPEC_GET(LED2_NODE, gpios);

static K_SEM_DEFINE(ble_init_ok, 0, 1);

bool btconnected = false;
bool btfirstadv = true;

#define BT_UUID_NUS_VAL BT_UUID_128_ENCODE(0x6E400001, 0xB5A3, 0xF393, 0xE0A9, 0xE50E24DCCA9E)
#define BT_UUID_NUS_RX_VAL BT_UUID_128_ENCODE(0x6E400002, 0xB5A3, 0xF393, 0xE0A9, 0xE50E24DCCA9E)
#define BT_UUID_NUS_TX_VAL BT_UUID_128_ENCODE(0x6E400003, 0xB5A3, 0xF393, 0xE0A9, 0xE50E24DCCA9E)
#define BT_UUID_NUS_PRESSURE_VAL BT_UUID_128_ENCODE(0x6E400004, 0xB5A3, 0xF393, 0xE0A9, 0xE50E24DCCA9E)
#define BT_UUID_NUS_TEMPERATURE_VAL BT_UUID_128_ENCODE(0x6E400005, 0xB5A3, 0xF393, 0xE0A9, 0xE50E24DCCA9E)
#define BT_UUID_LNS_LS_VAL 0x2A67
#define BT_UUID_LNS_PQ_VAL 0x2A69
#define BT_UUID_CUS_VAL 0x2000
#define BT_UUID_CUS_WRITE_VAL 0x2001
#define BT_UUID_CUS_SAT_VAL 0x2002
#define BT_UUID_CUS_CALIBRATE_VAL 0x2003

static uint8_t write_cus_buf[2];
uint8_t gpsmode = 1;

static ssize_t read_u32(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                        void *buf, uint16_t len, uint16_t offset);
static ssize_t read_s16(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                        void *buf, uint16_t len, uint16_t offset);
static ssize_t on_write_rx(struct bt_conn *conn, const struct bt_gatt_attr *attr, const void *buf, uint16_t len, uint16_t offset, uint8_t flags);
static ssize_t on_read_rx(struct bt_conn *conn, const struct bt_gatt_attr *attr, void *buf, uint16_t len, uint16_t offset);
static void nus_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value);
static void nus_ccc_cfg_changed2(const struct bt_gatt_attr *attr, uint16_t value);

static uint8_t nus_rx[26];
static uint8_t nus_tx_started;
static K_FIFO_DEFINE(fifo_btsendhisto);
struct btsendhisto_t {
        void *fifo_reserved;
        uint8_t type;
        uint32_t from;
};

enum captor_type { BATTERY, TEMPERATURE, PRESSURE };

struct s_captor_params {
  uint8_t wait_adv;
  uint8_t wait_record;
};

struct s_captors {
  uint8_t captor;
  struct s_captor_params params;
};

struct s_captors captors_params[] = {
  { TEMPERATURE, {1, 3}},
  { PRESSURE, {6, 12}},
  { BATTERY, {12, 12}},
};

struct s_captors captors_state[] = {
  { TEMPERATURE, {0, 0}},
  { PRESSURE, {0, 0}},
  { BATTERY, {0, 0}},
};

int16_t temperature;
uint32_t pressure;

bool findifadv(uint8_t captor)
{
  for(int i = 0; i < sizeof(captors_params) / sizeof(captors_params[0]); i++)
  {
    if(captors_params[i].captor == captor)
    {
      captors_state[i].params.wait_adv++;
      if(captors_state[i].params.wait_adv == captors_params[i].params.wait_adv)
      {
        captors_state[i].params.wait_adv = 0;
        return true;
      } else {
        return false;
      }
    }
  }
  return false;
}

bool findifrecord(uint8_t captor)
{
  for(int i = 0; i < sizeof(captors_params) / sizeof(captors_params[0]); i++)
  {
    if(captors_params[i].captor == captor)
    {
      captors_state[i].params.wait_record++;
      if(captors_state[i].params.wait_record == captors_params[i].params.wait_record)
      {
        captors_state[i].params.wait_record = 0;
        return true;
      } else {
        return false;
      }
    }
  }
  return false;
}

void reset_adv_state()
{
  for(int i = 0; i < sizeof(captors_state) / sizeof(captors_state[0]); i++)
  {
    captors_state[i].params.wait_adv = 0;
  }

}

void temperature_notify(struct sensor_value *sval);
void pressure_notify(struct sensor_value *sval);
//void humidity_notify(struct sensor_value *sval);
void ruuvi_advertise();

void update_captors(bool force)
{
  LOG_INF("Update Sensors");
  bool temperature_toadv = false;
  bool temperature_torecord = false;
  bool pressure_toadv = false;
  bool pressure_torecord = false;
  bool battery_toadv = false;
  bool battery_torecord = false;

  bool ppressure = false;
  bool ptemperature = false;

  struct sensor_value sval;
  
  if(btconnected == false)
  {
    temperature_toadv = findifadv(TEMPERATURE);
    pressure_toadv = findifadv(PRESSURE);
    battery_toadv = findifadv(BATTERY);
  }
  if(force == false)
  {
    temperature_torecord = findifrecord(TEMPERATURE);
    pressure_torecord = findifrecord(PRESSURE);
    battery_torecord = findifrecord(BATTERY);
  }
  
  if(temperature_torecord == true || temperature_toadv == true || force == true)
  {
    ptemperature = true;
  }
  if(pressure_torecord == true || pressure_toadv == true || force == true)
  {
    ppressure = true;
  }
  if(ptemperature == true || ppressure == true) 
  {
    probe_bme280(ptemperature, false, ppressure);
  }

  if(temperature_toadv == true || force == true)
  {
    LOG_INF("ADV Temperature");
    sval = get_temperature();
    if(sval.val1 != 0 || sval.val2 != 0)
    {
      //temperature_notify(&sval);
      ruuvi_updateTemperature(&sval);
      ruuvi_advertise();
    }
  }
  if(temperature_torecord == true && force == false)
  {
    LOG_INF("Record Temperature");
    sval = get_temperature();
    if(sval.val1 != 0 || sval.val2 != 0)
    {
      addRecord(TEMPERATURE, (sval.val1*100)+(int16_t)round(sval.val2/10000));
    }
  }
  if(pressure_toadv == true || force == true)
  {
    LOG_INF("ADV Pressure");
    sval = get_pressure();
    if(sval.val1 != 0 || sval.val2 != 0)
    {
      //pressure_notify(&sval);
      ruuvi_updatePressure(&sval);
      ruuvi_advertise();
    }
  }
  if(pressure_torecord == true && force == false)
  {
    LOG_INF("Record Pressure");
    sval = get_pressure();
    if(sval.val1 != 0 || sval.val2 != 0)
    {
      addRecord(PRESSURE, (sval.val1*100)+(int16_t)round(sval.val2/10000));
    }
  }
  if(battery_toadv == true || force == true)
  {
    LOG_INF("ADV Battery");
    uint16_t battery_millivolt = 0;
	  uint8_t battery_percentage = 0;

	  battery_get_millivolt(&battery_millivolt);
	  battery_get_percentage(&battery_percentage, battery_millivolt);

    ruuvi_updateBattery(battery_millivolt);
    ruuvi_advertise();

    bt_bas_set_battery_level(battery_percentage);
  }
}

static void work_captors_force_handler(struct k_work *work)
{
  update_captors(true);
}
K_WORK_DEFINE(work_captors_force, work_captors_force_handler);


static void work_captors_handler(struct k_work *work)
{
  update_captors(false);
}
K_WORK_DEFINE(work_captors, work_captors_handler);

void timer_captors_handler(struct k_timer *stm)
{
  k_work_submit(&work_captors);
}
K_TIMER_DEFINE(timer_captors, timer_captors_handler, NULL);

static ssize_t on_write_cus(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                            const void *buf, uint16_t len, uint16_t offset, uint8_t flags);

#define BT_UUID_LNS_LS BT_UUID_DECLARE_16(BT_UUID_LNS_LS_VAL)
#define BT_UUID_LNS_PQ BT_UUID_DECLARE_16(BT_UUID_LNS_PQ_VAL)
#define BT_UUID_CUS BT_UUID_DECLARE_16(BT_UUID_CUS_VAL)
#define BT_UUID_CUS_WRITE BT_UUID_DECLARE_16(BT_UUID_CUS_WRITE_VAL)
#define BT_UUID_CUS_SAT BT_UUID_DECLARE_16(BT_UUID_CUS_SAT_VAL)
#define BT_UUID_CUS_CALIBRATE BT_UUID_DECLARE_16(BT_UUID_CUS_CALIBRATE_VAL)


BT_GATT_SERVICE_DEFINE(lns_svc,
                       BT_GATT_PRIMARY_SERVICE(BT_UUID_LNS),
                       BT_GATT_CHARACTERISTIC(BT_UUID_LNS_LS,
                                              BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_NONE,
                                              NULL, NULL, NULL),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       BT_GATT_CHARACTERISTIC(BT_UUID_LNS_PQ,
                                              BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_NONE,
                                              NULL, NULL, NULL),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       );

BT_GATT_SERVICE_DEFINE(cus_svc,
                       BT_GATT_PRIMARY_SERVICE(BT_UUID_CUS),
                       BT_GATT_CHARACTERISTIC(BT_UUID_CUS_WRITE,
                                              BT_GATT_CHRC_WRITE | BT_GATT_CHRC_WRITE_WITHOUT_RESP,
                                              BT_GATT_PERM_WRITE,
                                              NULL, on_write_cus, &write_cus_buf),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       BT_GATT_CHARACTERISTIC(BT_UUID_CUS_SAT,
                                              BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_NONE,
                                              NULL, NULL, NULL),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       );



BT_GATT_SERVICE_DEFINE(ess_svc,
                       BT_GATT_PRIMARY_SERVICE(BT_UUID_ESS),
                       BT_GATT_CHARACTERISTIC(BT_UUID_TEMPERATURE,
                                              BT_GATT_CHRC_READ | BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_READ,
                                              read_s16, NULL, &temperature),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       BT_GATT_CHARACTERISTIC(BT_UUID_PRESSURE,
                                              BT_GATT_CHRC_READ | BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_READ,
                                              read_u32, NULL, &pressure),
                       BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       );

#define BT_UUID_NUS BT_UUID_DECLARE_128(BT_UUID_NUS_VAL)
#define BT_UUID_NUS_RX BT_UUID_DECLARE_128(BT_UUID_NUS_RX_VAL)
#define BT_UUID_NUS_TX BT_UUID_DECLARE_128(BT_UUID_NUS_TX_VAL)
#define BT_UUID_NUS_PRESSURE BT_UUID_DECLARE_128(BT_UUID_NUS_PRESSURE_VAL)
#define BT_UUID_NUS_TEMPERATURE BT_UUID_DECLARE_128(BT_UUID_NUS_TEMPERATURE_VAL)

BT_GATT_SERVICE_DEFINE(nus_svc,
                       BT_GATT_PRIMARY_SERVICE(BT_UUID_NUS),
    /* RX */
    BT_GATT_CHARACTERISTIC(BT_UUID_NUS_RX,
                           BT_GATT_CHRC_WRITE | BT_GATT_CHRC_WRITE_WITHOUT_RESP,
                           BT_GATT_PERM_READ  | BT_GATT_PERM_WRITE,
                           on_read_rx, on_write_rx, &nus_rx),
    /* TX */
    BT_GATT_CHARACTERISTIC(BT_UUID_NUS_TX,
                           BT_GATT_CHRC_NOTIFY,
                           BT_GATT_PERM_NONE,
                           NULL, NULL, NULL),
    BT_GATT_CCC(nus_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       /* Pressure */
                       BT_GATT_CHARACTERISTIC(BT_UUID_NUS_PRESSURE,
                                              BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_NONE,
                                              NULL, NULL, NULL),
                       BT_GATT_CCC(nus_ccc_cfg_changed2, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
                       /* Temperature */
                       BT_GATT_CHARACTERISTIC(BT_UUID_NUS_TEMPERATURE,
                                              BT_GATT_CHRC_NOTIFY,
                                              BT_GATT_PERM_NONE,
                                              NULL, NULL, NULL),
                       BT_GATT_CCC(nus_ccc_cfg_changed2, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE)
                       );


static const struct bt_data sd[] = {
  BT_DATA_BYTES(BT_DATA_UUID16_ALL,
                BT_UUID_16_ENCODE(BT_UUID_BAS_VAL),
                BT_UUID_16_ENCODE(BT_UUID_DIS_VAL),
                BT_UUID_16_ENCODE(BT_UUID_ESS_VAL),
                BT_UUID_16_ENCODE(BT_UUID_LNS_VAL),
                BT_UUID_16_ENCODE(BT_UUID_CUS_VAL)),
  BT_DATA_BYTES(BT_DATA_UUID128_ALL,
                BT_UUID_NUS_VAL),
};

static void nus_ccc_cfg_changed2(const struct bt_gatt_attr *attr, uint16_t value)
{
  LOG_WRN("%s:  handle: %d value: %d", __func__, attr->handle, value);
  //nus_tx_started = (value == BT_GATT_CCC_NOTIFY) ? 1 : 0;
}

static ssize_t read_u32(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                        void *buf, uint16_t len, uint16_t offset)
{
  const uint32_t *u32 = attr->user_data;
  uint32_t value = sys_cpu_to_le32(*u32);

  return bt_gatt_attr_read(conn, attr, buf, len, offset, &value,
                           sizeof(value));
}

static ssize_t read_s16(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                        void *buf, uint16_t len, uint16_t offset)
{
  const int16_t *s16 = attr->user_data;
  int16_t value = sys_cpu_to_le16(*s16);

  return bt_gatt_attr_read(conn, attr, buf, len, offset, &value,
                           sizeof(value));
}

static void work_notify_bme280_handler(struct k_work *work)
{
  struct sensor_value sval;

  probe_bme280(true, false, true);
  LOG_INF("Notify Pressure");
  sval = get_pressure();
  if(sval.val1 != 0 || sval.val2 != 0)
  {
    pressure_notify(&sval);
  }
  LOG_INF("Notify Temperature");
  sval = get_temperature();
  if(sval.val1 != 0 || sval.val2 != 0)
  {
    temperature_notify(&sval);
  }

}
K_WORK_DEFINE(work_notify_bme280, work_notify_bme280_handler);

void timer_notify_bme280_handler(struct k_timer *stm)
{
  k_work_submit(&work_notify_bme280);
}
K_TIMER_DEFINE(timer_notify_bme280, timer_notify_bme280_handler, NULL);

static void connected(struct bt_conn *conn, uint8_t err)
{
  if (err) {
    printk("Connection failed (err 0x%02x)\n", err);
  } else {
    printk("Connected\n");
    btconnected = true;
    k_timer_start(&timer_notify_bme280, K_NO_WAIT, K_SECONDS(10));
    gpsPower(true);
  }
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
  printk("Disconnected (reason 0x%02x)\n", reason);
  btconnected = false;
  k_timer_stop(&timer_notify_bme280);
  reset_adv_state();
  update_captors(true);
  gpsPower(false);
}


BT_CONN_CB_DEFINE(conn_callbacks) = {
  .connected = connected,
  .disconnected = disconnected,
};

static void auth_passkey_display(struct bt_conn *conn, unsigned int passkey)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Passkey for %s: %06u", addr, passkey);
}

static void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing cancelled: %s", addr);
}

static struct bt_conn_auth_cb auth_cb_display = {
	.passkey_display = auth_passkey_display,
	.cancel = auth_cancel,
  .pairing_confirm = NULL,
 // .pairing_complete = pairing_complete,
 // .pairing_failed = pairing_failed
};

static ssize_t on_write_rx(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                           const void *buf, uint16_t len, uint16_t offset, uint8_t flags)
{
    uint8_t * value = attr->user_data;

    if (offset + len > sizeof(nus_rx)) {
        LOG_ERR("%s: INVALID OFFSET: offset(%d) + len(%d) > sizeof(nus_rx) %d ",
                __func__, offset, len, sizeof(nus_rx));
        return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
    }

    memcpy(value + offset, buf, len);

    if(len == 11)
    {
        struct btsendhisto_t bthisto;
        uint32_t ts;
        ts = value[10] + (value[9] << 8) + (value[8] << 16) + (value[7] << 24);

        LOG_INF("Send histo from %d header %d %d %d", ts, value[0], value[1], value[2]);
        //btsendhisto(ts, value[0]);
        bthisto.from = ts;
        bthisto.type = value[0];
        size_t size = sizeof(struct btsendhisto_t);
        char *mem_ptr = k_malloc(size);
        memcpy(mem_ptr, &bthisto, size);
        //bt_gatt_notify_uuid(NULL, BT_UUID_NUS_TX, nus_svc.attrs, &bttosend, sizeof(bttosend));
        k_fifo_put(&fifo_btsendhisto, mem_ptr);
        LOG_INF("End send histo");
    }

    return len;
}

static ssize_t on_read_rx(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                            void *buf, uint16_t len, uint16_t offset)
{
    LOG_INF("%s", __func__);
    return bt_gatt_attr_read(conn, attr, buf, len, offset, &nus_rx, sizeof(nus_rx));
}

static void nus_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value)
{
    LOG_WRN("%s: value: %d", __func__, value);

    nus_tx_started = (value == BT_GATT_CCC_NOTIFY) ? 1 : 0;
}

void btsendhisto_thread()
{
  for (;;) {
    struct btsendhisto_t *request = k_fifo_get(&fifo_btsendhisto,
                                                     K_FOREVER);
    uint8_t bttosend[11];
    bttosend[0] = request->type;
    bttosend[2] = 0x10;
    for(int i=0;i<SIZE_DB;i++)
    {
        struct s_record r = database.db[i];
        if(r.timestamp > request->from && ((request->type == 0x30 && r.type == TEMPERATURE) || (request->type == 0x32 && r.type == PRESSURE) || request->type == 0x3a))
        {
            bttosend[6] = r.timestamp;
            bttosend[5] = r.timestamp >> 8;
            bttosend[4] = r.timestamp >> 16;
            bttosend[3] = r.timestamp >> 24;
            int32_t s32 = (int32_t) r.value;
            if(r.type == TEMPERATURE)
            {
                bttosend[1] = 0x30;
            } else if(r.type == PRESSURE)
            {
                bttosend[1] = 0x32;
                s32 = s32*10;
            }
            bttosend[10] = s32;
            bttosend[9] = s32 >> 8;
            bttosend[8] = s32 >> 16;
            bttosend[7] = s32 >> 24;
            LOG_INF("SRec %d ts %u type %d value %d", i, r.timestamp, r.type, r.value);
            bt_gatt_notify_uuid(NULL, BT_UUID_NUS_TX, nus_svc.attrs, &bttosend, sizeof(bttosend));
            k_msleep(50);
      }
    }
    bttosend[1] = request->type;
    bttosend[3] = 0xff;
    bttosend[4] = 0xff;
    bttosend[5] = 0xff;
    bttosend[6] = 0xff;
    bttosend[7] = 0xff;
    bttosend[8] = 0xff;
    bttosend[9] = 0xff;
    bttosend[10] = 0xff;
    bt_gatt_notify_uuid(NULL, BT_UUID_NUS_TX, nus_svc.attrs, &bttosend, sizeof(bttosend));

    k_free(request);
    }

}
K_THREAD_DEFINE(btsendhisto_thread_id, STACKSIZE, btsendhisto_thread, NULL, NULL,
                NULL, PRIORITY, 0, 0);


static ssize_t on_write_cus(struct bt_conn *conn, const struct bt_gatt_attr *attr,
                            const void *buf, uint16_t len, uint16_t offset, uint8_t flags)
{
  int8_t * value = attr->user_data;

  if (offset + len > sizeof(write_cus_buf)) {
    LOG_ERR("%s: INVALID OFFSET: offset(%d) + len(%d) > sizeof(auth_buf) %d ",
            __func__, offset, len, sizeof(write_cus_buf));
    return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
  }

  memcpy(value + offset, buf, len);

  if(len == 2)
  {
    if(value[1] & 0x01) {
      LOG_INF("Mode 1");
      gpsBoost(false);
      gpsmode = 1;
    } else if(value[1] & 0x02) {
      LOG_INF("Mode 2");
      gpsmode = 2;
      gpsBoost(true);
    } else if(value[1] & 0x04) {
      LOG_INF("Mode 3");
      gpsmode = 3;
      gpsBoost(true);

    }
  }
  return len;
}

void ruuvi_advertise()
{
  //LOG_HEXDUMP_INF(ruuvi_getPayload(), 16, "Ruuvi Payload:");
  struct bt_data ad[] = {
    BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
    BT_DATA(BT_DATA_MANUFACTURER_DATA, ruuvi_getPayload(), 16),
    BT_DATA(BT_DATA_NAME_COMPLETE, CONFIG_BT_DEVICE_NAME, sizeof(CONFIG_BT_DEVICE_NAME) - 1)
  };
  if(btfirstadv == true)
  {
    bt_le_adv_start(BT_LE_ADV_PARAM(BT_LE_ADV_OPT_CONNECTABLE, BT_GAP_ADV_FAST_INT_MIN_2, BT_GAP_ADV_FAST_INT_MAX_2, NULL), ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
    btfirstadv = false;
  } else {
    bt_le_adv_update_data(ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
  }
}

void temperature_notify(struct sensor_value *sval)
{
  int16_t tnotifynus, tnotifyess;

  tnotifynus = sys_cpu_to_be16((sval->val1*10)+(int16_t)round(sval->val2/100000));
  bt_gatt_notify_uuid(NULL, BT_UUID_NUS_TEMPERATURE, nus_svc.attrs, &tnotifynus, sizeof(tnotifynus));

  temperature = (sval->val1*100)+(int16_t)round(sval->val2/10000);
  tnotifyess = sys_cpu_to_le16(temperature);
  bt_gatt_notify_uuid(NULL, BT_UUID_TEMPERATURE, ess_svc.attrs, &tnotifyess, sizeof(tnotifyess));
}

void pressure_notify(struct sensor_value *sval)
{
  uint32_t pnotifynus, pnotifyess;

  pnotifynus = sys_cpu_to_be32((sval->val1*1000)+(int16_t)round(sval->val2/1000));
  bt_gatt_notify_uuid(NULL, BT_UUID_NUS_PRESSURE, nus_svc.attrs, &pnotifynus, sizeof(pnotifynus));

  pressure = (sval->val1*10000)+(int16_t)round(sval->val2/100);
  pnotifyess = sys_cpu_to_le32(pressure);
  bt_gatt_notify_uuid(NULL, BT_UUID_PRESSURE, ess_svc.attrs, &pnotifyess, sizeof(pnotifyess));
}

/* void humidity_notify(struct sensor_value *sval)
{
    uint16_t hnotifyess;

    humidity = (sval->val1*100)+(uint16_t)round(sval->val2/10000);
    hnotifyess = sys_cpu_to_le16(humidity);
    bt_gatt_notify_uuid(NULL, BT_UUID_HUMIDITY, ess_svc.attrs, &hnotifyess, sizeof(hnotifyess));
} */


void publish_bt_thread()
{
  k_sem_take(&ble_init_ok, K_FOREVER);
  for (;;) {
    k_sem_take(&gps_data_tosend, K_FOREVER);
    while(k_fifo_is_empty(&fifo_gps_data) == 0) {
      struct gps_data_t *request = k_fifo_get(&fifo_gps_data, K_NO_WAIT);
      uint8_t btgpsls[20];
      memcpy(&btgpsls[0], &(request->speed), 2 );
      memcpy(&btgpsls[2], &(request->latitude), 4 );
      memcpy(&btgpsls[6], &(request->longitude), 4 );
      memcpy(&btgpsls[10], &(request->acc_h), 4 );
      memcpy(&btgpsls[14], &(request->altitude), 4 );
      memcpy(&btgpsls[18], &(request->course), 2 );
      LOG_INF("Notify GPS Data LS");
      bt_gatt_notify_uuid(NULL, BT_UUID_LNS_LS, lns_svc.attrs, &btgpsls, sizeof(btgpsls));
      if(gpsmode == 3) {
        uint8_t btgpsverb[14];
        uint8_t hdop=(uint8_t) round(request->dop_h/2);
        uint8_t vdop=(uint8_t) round(request->dop_v/2);
        memcpy(&btgpsverb[0], &(request->acc_h), 4 );
        memcpy(&btgpsverb[4], &(request->acc_v), 4 );
        memcpy(&btgpsverb[8], &hdop, 1 );
        memcpy(&btgpsverb[9], &vdop, 1 );
        memcpy(&btgpsverb[10], &(request->altitude), 4 );
        LOG_INF("Notify GPS Data PQ");
        bt_gatt_notify_uuid(NULL, BT_UUID_LNS_PQ, lns_svc.attrs, &btgpsverb, sizeof(btgpsverb));
      }
      k_free(request);
    }
    while(k_fifo_is_empty(&fifo_gps_sat_data) == 0) {
      struct gps_sat_data_t *request = k_fifo_get(&fifo_gps_sat_data, K_NO_WAIT);
      if(gpsmode == 3) {
        uint8_t btgpssat[] = {request->gps, request->glonass, request->galileo};
        LOG_INF("Notify GPS Sat");
        bt_gatt_notify_uuid(NULL, BT_UUID_CUS_SAT, cus_svc.attrs, &btgpssat, sizeof(btgpssat));
      }
      k_free(request);
    }
  }
}
K_THREAD_DEFINE(publish_bt_id, STACKSIZE, publish_bt_thread, NULL, NULL, NULL, PRIORITY, 0, 0);

void led_init()
{
  gpio_pin_configure_dt(&led0, GPIO_OUTPUT_INACTIVE);
  gpio_pin_configure_dt(&led1, GPIO_OUTPUT_INACTIVE);
  gpio_pin_configure_dt(&led2, GPIO_OUTPUT_INACTIVE);
}

void led_test()
{
  gpio_pin_toggle_dt(&led0);
  k_msleep(350);
  gpio_pin_toggle_dt(&led0);
  gpio_pin_toggle_dt(&led1);
  k_msleep(350);
  gpio_pin_toggle_dt(&led1);
  gpio_pin_toggle_dt(&led2);
  k_msleep(350);
  gpio_pin_toggle_dt(&led2);
}

int main(void)
{
  LOG_INF("Hello main");

  led_init();
  led_test();

  k_sem_take(&gps_init_ok, K_FOREVER);

  int err = bt_enable(NULL);

  if (err) {
    printk("Bluetooth init failed (err %d)\n", err);
    return -1;
  }

  unsigned int passkey = 6768;
  bt_passkey_set(passkey);
  bt_conn_auth_cb_register(&auth_cb_display);

  k_sem_give(&ble_init_ok);

  battery_init();
  init_bme280();

  k_work_submit(&work_captors_force);
  k_timer_start(&timer_captors, K_SECONDS(300), K_SECONDS(300));


  //k_msleep(7500);
  //gpsPower(false);
  return 0;
}
